/**
 * Javier Abelln. 6 Abril 2006
 * 
 * Librera grfica
 */

package chuidiang.graficos;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import java.util.LinkedList;

import chuidiang.matematicas.CartesianasPolares;

/**
 * Clase encargada de transformar coordenadas de usuario en pixels y viceversa
 * en grficos con coordenadas polares.<br>
 */
public class EscalaGraficaPolar implements InterfaceEscalaGrafica
{
	// ~ Variables de instancia
	// --------------------------------------------------

	/** Los polares se convierten a cartesiana y se delega en esta clase */
	protected EscalaGraficaCartesiana escala = new EscalaGraficaCartesiana();

	/** Observadores de cambios en los extremos del grfico */
	private LinkedList<ObservadorEscalaGrafica> observadores = new LinkedList<ObservadorEscalaGrafica>();

	/** Clase para ampliar y reducir el grfico polar */
	private ZoomEscalaGraficaPolar zoom;

	// ~ Constructores
	// -----------------------------------------------------------

	/**
	 * Crea un nuevo objeto de la clase EscalaGraficaPolar.<br>
	 */
	public EscalaGraficaPolar()
	{
		zoom = new ZoomEscalaGraficaPolar(this);
	}

	// ~ Metodos
	// -----------------------------------------------------------------

	/**
	 * Devuelve un rectangulo con los extremos cartesianos del grfico.<br>
	 * 
	 * @return Extremos cartesianos del grfico.<br>
	 */
	public Rectangle2D getExtremosCartesianos()
	{
		return escala.dameExtremos();
	}

	/**
	 * Devuelve una clase adecuada para hacer zoom en un grfico con una
	 * EscalaGraficaPolar
	 * 
	 * @return Clase para hacer zoom.<br>
	 */
	public InterfaceZoomable getZoomable()
	{
		return zoom;
	}

	/**
	 * Aade un observador a la lista d eobservadores para avisarle cuando haya
	 * cambio en los extremos del grfico.<br>
	 * 
	 * @param unObservador
	 *            El observador.<br>
	 */
	public void anhadeObservador(ObservadorEscalaGrafica unObservador)
	{
		observadores.add(unObservador);
	}

	/**
	 * Devuelve las coordenadas de usuario de un punto expresado en pixels.<br>
	 * til para transformar coordenadas de un evento de ratn en coordenadas de
	 * usuario.
	 * 
	 * @param punto
	 *            Punto con coordenadas en pixels.<br>
	 * @return Punto con coordenadas de usuario.<br>
	 */
	public Point2D dameCoordenadaUsuario(Point2D punto)
	{
		Point2D pUsuario = escala.dameCoordenadaUsuario(punto);
		Point2D pUsuarioPolar = new Point2D.Double();
		pUsuarioPolar.setLocation(CartesianasPolares.dameModulo(
				pUsuario.getX(), pUsuario.getY()), CartesianasPolares
				.dameAngulo(pUsuario.getX(), pUsuario.getY()));

		return pUsuarioPolar;
	}

	/**
	 * Devuelve el area de dibujo en coordenadas polares de usuario.<br>
	 * 
	 * @return Rectangulo de dibujo.<br>
	 */
	public Rectangle2D dameExtremos()
	{
		Rectangle2D rCartesiano = escala.dameExtremos();
		Rectangle2D rPolar = new Rectangle2D.Double();
		double rMin = CartesianasPolares.dameModulo(rCartesiano.getX(),
				rCartesiano.getY());
		double tMin = CartesianasPolares.dameAngulo(rCartesiano.getX(),
				rCartesiano.getY());
		rPolar.setRect(rMin, tMin, rCartesiano.getWidth(), rCartesiano
				.getHeight());

		return rPolar;
	}

	/**
	 * Elimina el observador que se le pasa de la lista de observadores.<br>
	 * Ya no le avisar de cambios en los lmites del dibujo
	 * 
	 * @param unObservador
	 *            El observador a eliminar.<br>
	 */
	public void eliminaObservador(ObservadorEscalaGrafica unObservador)
	{
		observadores.remove(unObservador);
	}

	/**
	 * Dibuja un arco.<br>
	 * El arco pertenece a una circunferencia que va inscrita en un rectngulo
	 * @param esquina
	 *            Esquina del rectngulo.<br>
	 * @param ancho
	 *            Ancho del rectngulo.<br>
	 * @param alto
	 *            Alto del rectngulo.<br>
	 * @param arcoInicial
	 *            Angulo inicial del arco, en radianes.<br>
	 * @param arcoFinal
	 *            Angulo final del arco, en radianes.<br>
	 * @param color
	 *            Color del arco.<br>
	 */
	public void pintaArco(Point2D centro, double ancho, double alto,
			double arcoInicial, double arcoFinal, Color color)
	{

		Point2D centroCartesiano = new Point2D.Double();
		centroCartesiano.setLocation(CartesianasPolares.dameX(centro.getX(),
				centro.getY()), CartesianasPolares.dameY(centro.getX(), centro
				.getY()));
		this.escala.pintaArco(centroCartesiano, ancho, alto, arcoInicial,
				arcoFinal, color);
	}

	/**
	 * Dibuja un arco relleno.<br>
	 * El arco pertenece a una circunferencia que va inscrita en un rectngulo
	 * @param esquina
	 *            Esquina del rectngulo.<br>
	 * @param ancho
	 *            Ancho del rectngulo.<br>
	 * @param alto
	 *            Alto del rectngulo.<br>
	 * @param arcoInicial
	 *            Angulo inicial del arco, en radianes.<br>
	 * @param arcoFinal
	 *            Angulo final del arco, en radianes.<br>
	 * @param color
	 *            Color del arco.<br>
	 */
	public void pintaArcoRelleno(Point2D centro, double ancho, double alto,
			double arcoInicial, double arcoFinal, Color color)
	{
		Point2D centroCartesiano = new Point2D.Double();
		centroCartesiano.setLocation(CartesianasPolares.dameX(centro.getX(),
				centro.getY()), CartesianasPolares.dameY(centro.getX(), centro
				.getY()));
		this.escala.pintaArcoRelleno(centroCartesiano, ancho, alto,
				arcoInicial, arcoFinal, color);
	}

	/**
	 * Dibuja un eje de las x (radio) en la coordenada y (con angulo determinado).<br>
	 * 
	 * @param y
	 *            Coordenada y para dibujar el eje.<br>
	 * @param color
	 *            Color del eje.<br>
	 */
	public void pintaEjeX(double y, Color color)
	{
		Rectangle2D extremos = dameExtremos();
		Point2D.Double[] eje = new Point2D.Double[2];
		eje[0] = new Point2D.Double(0, y);
		eje[1] = new Point2D.Double(extremos.getWidth(), y);
		pintaPoliLinea(eje, color);
	}

	/**
	 * Pinta el eje de los angulos, es decir un circulo del radio x del centro
	 * que se pasa.<br>
	 * El color
	 * 
	 * @param x
	 *            Radio para dibujar el crculo.<br>
	 * @param color
	 *            Color del crculo.<br>
	 */
	public void pintaEjeY(double x, Color color)
	{
		Point2D centro = new Point2D.Double(
				CartesianasPolares.dameModulo(x, x), 5 * Math.PI / 4.0);
		pintaArco(centro, 2 * x, 2 * x, 0.0, 360.0, color);
	}

	/**
	 * Dibuja una polgono abierto, uniendo consecutivamente los puntos de
	 * puntos.<br>
	 * 
	 * @param puntos
	 *            Puntos del polgono.<br>
	 * @param color
	 *            Color para el polgono.<br>
	 */
	public void pintaPoliLinea(Point2D[] puntos, Color color)
	{
		Point2D[] pCartesianos = new Point2D.Double[puntos.length];

		for (int i = 0; i < puntos.length; i++)
		{
			pCartesianos[i] = new Point2D.Double(CartesianasPolares.dameX(
					puntos[i].getX(), puntos[i].getY()), CartesianasPolares
					.dameY(puntos[i].getX(), puntos[i].getY()));
		}

		escala.pintaPoliLinea(pCartesianos, color);
	}

	/**
	 * Dibuja los puntos que se le pasan.<br>
	 * 
	 * @param puntos
	 *            Los puntos.<br>
	 * @param color
	 *            Color de cada uno de los puntos.<br>
	 */
	public void pintaPuntos(Point2D[] puntos, Color[] color)
	{
		Point2D[] pCartesianos = new Point2D.Double[puntos.length];

		for (int i = 0; i < puntos.length; i++)
		{
			pCartesianos[i] = new Point2D.Double(CartesianasPolares.dameX(
					puntos[i].getX(), puntos[i].getY()), CartesianasPolares
					.dameY(puntos[i].getX(), puntos[i].getY()));
		}

		escala.pintaPuntos(pCartesianos, color);
	}

	/**
	 * Pinta puntos de un mismo color.<br>
	 * 
	 * @param puntos
	 *            Los puntos.<br>
	 * @param color
	 *            Color para los puntos.<br>
	 */
	public void pintaPuntos(Point2D[] puntos, Color color)
	{
		Point2D[] pCartesianos = new Point2D.Double[puntos.length];

		for (int i = 0; i < puntos.length; i++)
		{
			pCartesianos[i] = new Point2D.Double(CartesianasPolares.dameX(
					puntos[i].getX(), puntos[i].getY()), CartesianasPolares
					.dameY(puntos[i].getX(), puntos[i].getY()));
		}

		escala.pintaPuntos(pCartesianos, color);
	}

	/**
	 * Pinta un rectngulo relleno.<br>
	 * 
	 * @param esquina
	 *            esquina inferior izquierda del rectngulo.<br>
	 * @param ancho
	 *            Ancho del rectngulo.<br>
	 * @param alto
	 *            Alto del rectngulo.<br>
	 * @param color
	 *            Color del rectngulo.<br>
	 */
	public void pintaRectanguloRelleno(Point2D esquina, double ancho,
			double alto, Color color)
	{
		// TODO Auto-generated method stub
	}

	/**
	 * Pinta un rectngulo relleno en 3D.<br>
	 * 
	 * @param esquina
	 *            Esquina inferior izquierda del rectngulo.<br>
	 * @param ancho
	 *            Ancho del rectngulo.<br>
	 * @param alto
	 *            Alto del rectngulo.<br>
	 * @param color
	 *            Color del rectngulo.<br>
	 * @param elevado
	 *            COMENTARIO.<br>
	 */
	public void pintaRectanguloRelleno3D(Point2D esquina, double ancho,
			double alto, Color color, boolean elevado)
	{
		// TODO Auto-generated method stub
	}

	/**
	 * Dibuja un texto en la posicin indicada.<br>
	 * 
	 * @param posicion
	 *            Posicin del texto.<br>
	 * @param texto
	 *            El texto.<br>
	 * @param color
	 *            Color.<br>
	 */
	public void pintaTexto(Point2D posicion, String texto, Color color)
	{
		Point2D.Double posicionTexto = new Point2D.Double();
		posicionTexto.setLocation(CartesianasPolares.dameX(posicion.getX(),
				posicion.getY()), CartesianasPolares.dameY(posicion.getX(),
				posicion.getY()));
		escala.pintaTexto(posicionTexto, texto, color);
	}

	/**
	 * Se le pasan los lmites para el grfico polar.<br>
	 * 
	 * @param rMin
	 *            Radio mnimo que se quiere dibujar.<br>
	 * @param tMin
	 *            Angulo mnimo en radianes.<br>
	 * @param rMax
	 *            Radio mximo.<br>
	 * @param tMax
	 *            Angulo mximo en radianes.<br>
	 */
	public void tomaExtremos(double rMin, double tMin, double rMax, double tMax)
	{
		double xMin = CartesianasPolares.dameX(rMin, tMin);
		double yMin = CartesianasPolares.dameY(rMin, tMin);
		double xMax = CartesianasPolares.dameX(rMax, tMax);
		double yMax = CartesianasPolares.dameY(rMax, tMax);
		this.escala.tomaExtremos(xMin, yMin, xMax, yMax);
		this.avisaObseradores(rMin, tMin, rMax, tMax);
	}

	/**
	 * Se le pasan los lmites cartesianos para dibujar.<br>
	 * Las unidades son las del radio y se supone que van en vertical y horizontal
	 * 
	 * @param xMin
	 *            x minimo.<br>
	 * @param yMin
	 *            y minimo.<br>
	 * @param xMax
	 *            x maximo.<br>
	 * @param yMax
	 *            y maximo.<br>
	 */
	public void tomaExtremosCartesianos(double xMin, double yMin, double xMax,
			double yMax)
	{
		this.escala.tomaExtremos(xMin, yMin, xMax, yMax);
		this.avisaObseradores(CartesianasPolares.dameModulo(xMin, yMin),
				CartesianasPolares.dameAngulo(xMin, yMin), CartesianasPolares
						.dameModulo(xMax, yMax), CartesianasPolares.dameAngulo(
						xMax, yMax));
	}

	/**
	 * Se le pasa el Graphics sobre el que va a dibujar, as como el
	 * alto y ancho en pixels del rea de dibujo.<br>
	 * 
	 * @param g
	 *            Graphics para dibujar.<br>
	 * @param ancho
	 *            Ancho en pixels del rea de dibujo.<br>
	 * @param alto
	 *            Alto en pixels del rea de dibujo.<br>
	 */
	public void tomaGraphics(Graphics g, int ancho, int alto)
	{
		escala.tomaGraphics(g, ancho, alto);
	}

	/**
	 * Avisa a los observadores de cambio en los lmites del grfico de que
	 * se ha producido un cambio en dichos lmites.<br>
	 * 
	 * @param xMin
	 *            Nueva x min.<br>
	 * @param yMin
	 *            Nueva y min.<br>
	 * @param xMax
	 *            Nueva x max.<br>
	 * @param yMax
	 *            Nueva y max.<br>
	 */
	private void avisaObseradores(double xMin, double yMin, double xMax,
			double yMax)
	{
		for (int i = 0; i < observadores.size(); i++)
		{
			((ObservadorEscalaGrafica) observadores.get(i)).tomaNuevosExtremos(
					xMin, yMin, xMax, yMax);
		}
	}

    /**
     * Dibuja la imagen que se le pasa en la posicin indicada.
     * 
     * @param posicion La posicin para dibujar la imgen
     * @param imagen La imagen a dibujar
     */
	public void pintaImagen(Point2D posicion, Image imagen)
	{
		if (posicion == null)
			return;
		Point2D.Double posicionImagen = new Point2D.Double();
		posicionImagen.setLocation(CartesianasPolares.dameX(posicion.getX(),
				posicion.getY()), CartesianasPolares.dameY(posicion.getX(),
				posicion.getY()));
		this.escala.pintaImagen(posicionImagen, imagen);
	}
}
